"use strict";

/*
 * Logging levels for `Database.setLoggingLevel()` function.
 * No other logging levels are supported.
 */

const LOGGING_LEVELS = {
  DEBUG: "debug",
  INFO: "info",
  ERROR: "err",
};

module.exports = LOGGING_LEVELS;
